/*
   This program creates a call graph describing how every function
   calls every other function.

   Change log:
   04/14/92 - created version 1.0 John Henckel
   11/18/92 - added codes for comments using "//"
            - added codes for double quotations
            - added codes for old-type function declarations, like:
                             a(b, c) int b; int c;
                             {
                               printf("b = %d, c = %d", b, c);
                             }
                                 by M. Takamiya (jl22373 at ymtvm6)
   01/18/93 - fixed problem with prev to avoid confusing function declares
              with old-type definitions.
            - fixed problem with "\"" strings and '"' and '\"' char.
            - removed comments from buffer when detected.
            - make string removal optional (can be disasterous).
            - make comment removal optional (can be disasterous).
            - fix "used" count of recursive functions.
            - right justify filename.
   01/18/93 - created version 1.1
   02/16/94 - enhanced /d to use the /h option
            - added option /b to make backwards invocation tree
            - also option /q (noisy).
   03/10/97 - added option /y to search for cycles.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#define mask 0xFF
#define chunk 100
#define tsize 35

/*--------------------------------------------------------------------
  This structure describes an entry-point name.
    used - number of times this function is used.
    uses - number of other functions this one uses.
    calls- number of other defined functions this one uses.
    file - index in epn of the filename
    def  - is this one defined in one of the files?
    done - was this one already printed?
    s    - name (truncated to 19 ch)
*/

typedef struct {
  int used,uses,file,calls;
  char def,done;
  char s[tsize+1];
} epn_t;

typedef struct {
  int from,to;
} con_t;

char buffer[mask+3];
epn_t *epn;
con_t *con;
char s[80];
int rc,maxepn=0,numepn=0;
int maxcon=0,numcon=0;
int curfile=0;
int just=39,back=0,noisy=1,cycl=0;
int dump=0,hide=0,xref=0,multi=0,strg=0,cmmt=0;   /* cmdline options */

/*
   These are the only characters that can be used in C identifiers
*/
letter(int ch)
{
  return (ch<='9' && ch >='0' || ch<='Z' && ch >='A' ||
          ch<='z' && ch >='a' || ch=='_');
}

addepn(char *name)
{
  int i;

  for (i=1; i<=numepn; ++i)
    if (!strcmp(epn[i].s,name))
      return i;

  if (numepn>=maxepn-1) {         /* grow */
    if (!maxepn) epn=NULL;
    maxepn += chunk;
    epn = realloc(epn,maxepn*sizeof *epn);
  }
  ++numepn;

  memset(epn+numepn,0,sizeof *epn);
  strncpy(epn[numepn].s,name,tsize);
  epn[numepn].s[tsize] = 0;
  /*
  epn[numepn].used = 0;
  epn[numepn].uses = 0;
  epn[numepn].file = 0;
  epn[numepn].def = 0;
  epn[numepn].done = 0; */

  return numepn;
}

addcon(int from, int to)
{
  int i;

  for (i=numcon-1; i>=0; --i)
    if (con[i].from != from) break;
    else if (con[i].to==to) return 0;

  if (numcon>=maxcon) {         /* grow */
    if (!maxcon) con=NULL;
    maxcon += chunk;
    con = realloc(con,maxcon*sizeof *con);
  }
  con[numcon].from = from;
  con[numcon].to = to;
  ++numcon;

  if (to!=from) epn[to].used++;      /* don't count self as a user */
  epn[from].uses++;

  return numcon;
}

keyword(char *name)
{
  if (!*name) return 1;                     /* is name empty */
  if (*name>='0' && *name<='9') return 1;   /* is the first letter numeric */
  if (!strcmp(name,"if"      )) return 1;
  if (!strcmp(name,"sizeof"   )) return 1;
  if (!strcmp(name,"return"  )) return 1;
  if (!strcmp(name,"for"     )) return 1;
  if (!strcmp(name,"while"   )) return 1;
  if (!strcmp(name,"else"    )) return 1;
  if (!strcmp(name,"switch"  )) return 1;
  return 0;
}

readfile(char *fname)
{
  int ch,level=0,curid=0,id;
  long n=0,i,j;
  FILE *f;

  f = fopen(fname,"r");
  if (f==NULL) return fprintf(stderr,"Unable to open '%s'.\n",fname);

  if (noisy) printf("%s\n",fname);

  while(1) {
    ch = fgetc(f);
    if (ch < 0) break;
    buffer[n & mask] = ch;
    n++;
    switch (ch) {
      case '(':               /* find identifier */
        for (i=2; i<mask && buffer[(n-i) & mask]<=' '; ++i);

        for (j=0; j<tsize && (j+i)<mask && letter(buffer[(n-i-j) & mask]); ++j)
          s[tsize-1-j] = buffer[(n-i-j) & mask];

        s[tsize]=0;                /* make null terminated */
        if (keyword(s+tsize-j)) break;
        id = addepn(s+tsize-j);
        if (level < 1) {
          curid = id;
          if (noisy) printf("   %s\n",s+tsize-j);
        }
        else if (curid) addcon(curid,id);
        break;
      case '{':
        ++level;
        if (level==1 && curid) {
          if (epn[curid].def) fprintf(stderr,
            "The function '%s' is defined in '%s' and '%s'.\n",
            epn[curid].s,epn[epn[curid].file].s,epn[curfile].s);
          epn[curid].def = 1;
          epn[curid].file = curfile;
        }
        break;
      case '}': --level;
        if (!level) curid = 0;
        break;
      case ';':
        /*--------------------------------------------------------------------
           At level 0, a ; could be the end of a function prototype or
           it could delimit the old-style parameter definitions.
        */
        if (!level && curid) {
          for (i=2; i<mask && buffer[(n-i) & mask]<=' '; ++i);
          if (i<mask && buffer[(n-i) & mask]==')') curid = 0;
        }
        break;
      case '*':                      /* could be a comment -- skip it */
        if (cmmt) break;
        if (buffer[(n-2) & mask]!='/') break;
        n-=2;
        ch = fgetc(f);
        do {
          while (ch >= 0 && ch != '*') ch=fgetc(f);
        } while (ch >= 0 && (ch=fgetc(f)) != '/');
        if (ch < 0)
          fprintf(stderr,"There are unmatched /* in %s\n",fname);
        break;
      case '/':
        if (cmmt) break;
        if (buffer[(n-2) & mask]!='/') break;
        n-=2;
        ch = fgetc(f);
        while (ch >= 0 && ch != '\n') ch=fgetc(f);
        break;
      case '"':
        if (strg) break;
        if (buffer[(n-2) & mask]=='\'') break;  /* handle '"' */
        if (buffer[(n-2) & mask]=='\\') break;  /* handle '\"' */
        ch = fgetc(f);
        while (ch >= 0 && ch != '"') {
          if (ch=='\\') ch=fgetc(f);     /* handle "\"\n"  */
          ch=fgetc(f);
        }
        if (ch < 0)
          fprintf(stderr,"There are unmatched \" in %s\n",fname);
        break;
    }
  }
  if (level>0) fprintf(stderr,"There are %d unmatched { in %s\n",level,fname);
  if (level<0) fprintf(stderr,"There are %d unmatched } in %s\n",-level,fname);

  fclose(f);
  return 0;
}

/*--------------------------------------------------------------------
   recursive function to print tree (or dump)
*/
printtree(int t,int indent)
{
  int i,j;

  if (epn[t].def==0 && hide) return 0;

  for (i=0; i<indent; ++i) printf("   ");

  if (epn[t].def==0)
    printf("%s\n",epn[t].s);
  else {
    printf("%s  ",epn[t].s);

    i=strlen(epn[t].s);
    j=strlen(epn[epn[t].file].s);
    i = just - indent*3 - i - j - 5;        /* right justify */
    for ( ; i>0; --i) printf(" ");

    if (epn[t].done || (xref && indent))
      return printf("+  %s\n",epn[epn[t].file].s);

    printf("-  %s\n",epn[epn[t].file].s);
  }
  epn[t].done = 1;

  if (back) {
    for (i=0; i<numcon; ++i)
      if (con[i].to==t) printtree(con[i].from,indent+1);
  }
  else
    for (i=0; i<numcon; ++i)
      if (con[i].from==t) printtree(con[i].to,indent+1);

  return 0;
}

/**/

/*---------------------------------------------------------------
   recursive function to find and print any cycles.
   returns number of cycles found.
*/
findcycle(int t, int top) {
  int i,j,f=0;
  for (i=0; i<numcon; ++i)
    if (con[i].from == t &&
        epn[j = con[i].to].uses) {
      if (j==top) {
        printf("CYCLE: ");
        while (j) {
          printf("%s ", epn[j].s);
          j = j==t ? 0 : epn[j].done;
        }
        printf("\n");
        ++f;
      }
      else if (j > top && j!=t && !epn[j].done) {
	epn[t].done = j;
        f+=findcycle(j,top);
	epn[t].done = 0;
      }
    }
  return f;
}


writeout()
{
  int i;

  for (i=0; i<numcon; ++i)          /* set the number of defined fns called */
    if (epn[con[i].to].def && con[i].to!=con[i].from)
      ++epn[con[i].from].calls;

  if (dump) {
    printf("!numtoken %d numcon %d\n",numepn,numcon);
    for (i=1; i<=numepn; ++i)
      if (!hide || epn[i].def)
      printf("%4d %-25s uses %d used %d file %d def %d done %d\n",i,epn[i].s,
         epn[i].uses,epn[i].used,epn[i].file,(int)epn[i].def,(int)epn[i].done);
    for (i=0; i<numcon; ++i)
      if (!hide || epn[con[i].from].def && epn[con[i].to].def)
      printf("%d calls %d\n",con[i].from,con[i].to);

    return 0;
  }
  if (cycl) {
    int f=0;
    printf("Searching for cycles...\n");
    for (i=1; i<=numepn; ++i)
      f+=findcycle(i,i);
    printf("Found %d cycles.\n",f);
    return 0;
  }
  printf("-------------------------------------------------------------------\n"
         "Key:\n"
         "   - = this is the first reference\n"
         "   + = this is the second reference\n"
         "     = not defined in any of the files\n");

  for (i=1; i<=numepn; ++i)
    if (back) {                 /* backwards tree, children=called-by */
      if (!epn[i].calls || xref)
      printtree(i,0);
    }
    else if (epn[i].def && (!epn[i].used || xref))      /* is it a root */
      printtree(i,0);

}

/***/

main(int argc, char *argv[], char *envp[])
{
  FILE *g;
  char fname[80];
  int i;

  if (argc<2 || argv[1][0]=='?') {
    printf(
    "CTREE by John Henckel, IBM Internal Use Only, version "__DATE__"\n"
    "This program creates a call graph describing who calls who in a C\n"
    "program.  This program does not have a proper parser, it only looks\n"
    ); printf(
    "at the characters { } and ( to determine (guess) what things are\n"
    "function definitions.  By default, anything in /* comments */ or\n"
    "following a // comment or enclosed in \"quotes\" is ignored.\n"
    ); printf(
    "Comments and strings are not always detected correctly, so this\n"
    "feature may be disabled.  If you use macros extensively, it may be\n"
    "helpful to use your compiler to preprocess the code before you run\n"
    ); printf(
    "ctree.  Maximum token size is %d chars.  To redirect the output,\n"
    "use >outfile on the command line.\n\n",tsize
    ); printf(
    "Parameters:  infile options\n"
    "Options:  /m = infile contains a list of filenames.\n"
    "          /h = hide functions not explicitly defined (e.g. strcpy).\n"
    "          /x = limit depth of tree to 1 (like an xref).\n"
    "          /b = make a backwards tree (called-by).\n"
    ); printf(
    "          /s = do not try to skip over string constants.\n"
    "          /c = do not try to skip over comments.\n"
    "          /jnn = set column to right justify filename (def=%d).\n",just
    ); printf(
    "          /q = do not print files and functions during parse.\n"
    "          /y = show circular function invocations.\n"
    "          /d = dump internal arrays.\n"
    ); return 0;
  }

  for (i=2; i<argc; ++i)
    if (!strcmp(argv[i],"/m")) multi=1; else
    if (!strcmp(argv[i],"/h")) hide =1; else
    if (!strcmp(argv[i],"/x")) xref =1; else
    if (!strcmp(argv[i],"/y")) cycl =1; else
    if (!strcmp(argv[i],"/b")) back =1; else
    if (!strcmp(argv[i],"/s")) strg =1; else
    if (!strcmp(argv[i],"/c")) cmmt =1; else
    if (!strcmp(argv[i],"/d")) dump =1; else
    if (!strcmp(argv[i],"/q")) noisy=0; else
    if (!memcmp(argv[i],"/j",2)) sscanf(argv[i]+2,"%d",&just); else
    fprintf(stderr,"unknown parameter '%s' ignored\n",argv[i]);

  if (multi) {
    g = fopen(argv[1],"r");
    while (1) {
      rc = fscanf(g,"%s",fname);
      if (rc==1) {
        curfile = addepn(fname);
        epn[curfile].calls=1;   /* prevent it from showing up in /b */
        readfile(fname);
      }
      else break;
    }
    fclose(g);
  }
  else {
    curfile = addepn(argv[1]);
    epn[curfile].calls=1;
    readfile(argv[1]);
  }

  writeout();

  return 0;
}
